@@ -21,6 +21,8 @@
 
 #include "zend_portability.h"
 
+typedef struct _zend_string zend_string;
+
 #define ZEND_MAP_PTR_KIND_PTR           0
 #define ZEND_MAP_PTR_KIND_PTR_OR_OFFSET 1
 
@@ -40,6 +42,9 @@
 #define ZEND_MAP_PTR_NEW(ptr) do { \
 		ZEND_MAP_PTR(ptr) = zend_map_ptr_new(); \
 	} while (0)
+#define ZEND_MAP_PTR_NEW_STATIC(ptr) do { \
+		ZEND_MAP_PTR(ptr) = zend_map_ptr_new_static(); \
+	} while (0)
 
 #if ZEND_MAP_PTR_KIND == ZEND_MAP_PTR_KIND_PTR_OR_OFFSET
 # define ZEND_MAP_PTR_NEW_OFFSET() \
@@ -51,7 +56,7 @@
 		ZEND_MAP_PTR_GET_IMM(ptr) : \
 		((void*)(ZEND_MAP_PTR(ptr)))))
 # define ZEND_MAP_PTR_GET_IMM(ptr) \
-	(*ZEND_MAP_PTR_OFFSET2PTR((uintptr_t)ZEND_MAP_PTR(ptr)))
+	(*ZEND_MAP_PTR_OFFSET2PTR((intptr_t)ZEND_MAP_PTR(ptr)))
 # define ZEND_MAP_PTR_SET(ptr, val) do { \
 		if (ZEND_MAP_PTR_IS_OFFSET(ptr)) { \
 			ZEND_MAP_PTR_SET_IMM(ptr, val); \
@@ -60,11 +65,14 @@
 		} \
 	} while (0)
 # define ZEND_MAP_PTR_SET_IMM(ptr, val) do { \
-		void **__p = ZEND_MAP_PTR_OFFSET2PTR((uintptr_t)ZEND_MAP_PTR(ptr)); \
+		void **__p = ZEND_MAP_PTR_OFFSET2PTR((intptr_t)ZEND_MAP_PTR(ptr)); \
 		*__p = (val); \
 	} while (0)
 # define ZEND_MAP_PTR_BIASED_BASE(real_base) \
-	((void*)(((uintptr_t)(real_base)) - 1))
+	((void*)(((uintptr_t)(real_base)) + zend_map_ptr_static_size * sizeof(void *) - 1))
+/* Note: chunked like: [8192..12287][4096..8191][0..4095] */
+#define ZEND_MAP_PTR_STATIC_NUM_TO_PTR(num) \
+	((void **)CG(map_ptr_real_base) + zend_map_ptr_static_size - ZEND_MM_ALIGNED_SIZE_EX((num) + 1, 4096) + ((num) & 4095))
 #else
 # error "Unknown ZEND_MAP_PTR_KIND"
 #endif
@@ -73,9 +81,13 @@ BEGIN_EXTERN_C()
 
 ZEND_API void  zend_map_ptr_reset(void);
 ZEND_API void *zend_map_ptr_new(void);
+ZEND_API void *zend_map_ptr_new_static(void);
 ZEND_API void  zend_map_ptr_extend(size_t last);
 ZEND_API void zend_alloc_ce_cache(zend_string *type_name);
 
+ZEND_API extern size_t zend_map_ptr_static_last;
+ZEND_API extern size_t zend_map_ptr_static_size;
+
 END_EXTERN_C()
 
 #endif /* ZEND_MAP_PTR_H */
